# -*- coding: utf-8 -*-



# This program is free software; you can redistribute it and/or modify

# it under the terms of the (LGPL) GNU Lesser General Public License as

# published by the Free Software Foundation; either version 3 of the

# License, or (at your option) any later version.

#

# This program is distributed in the hope that it will be useful,

# but WITHOUT ANY WARRANTY; without even the implied warranty of

# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

# GNU Library Lesser General Public License for more details at

# ( http://www.gnu.org/licenses/lgpl.html ).

#

# You should have received a copy of the GNU Lesser General Public License

# along with this program; if not, write to the Free Software

# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )



"""

Unit tests related to Suds Python library reply processing.



Implemented using the 'pytest' testing framework.



"""



if __name__ == "__main__":

    from . import __init__

    __init__.runUsingPyTest(globals())





import suds

import tests



import pytest



import http.client

import re

import xml.sax





def test_ACCEPTED_and_NO_CONTENT_status_reported_as_None_with_faults():

    client = tests.client_from_wsdl(_wsdl__simple, faults=True)

    f = lambda r, s : client.service.f(__inject={"reply":suds.byte_str(r),

        "status":s})

    assert f("", None) is None

    pytest.raises(Exception, f, "", http.client.INTERNAL_SERVER_ERROR)

    assert f("", http.client.ACCEPTED) is None

    assert f("", http.client.NO_CONTENT) is None

    assert f("bla-bla", http.client.ACCEPTED) is None

    assert f("bla-bla", http.client.NO_CONTENT) is None





def test_ACCEPTED_and_NO_CONTENT_status_reported_as_None_without_faults():

    client = tests.client_from_wsdl(_wsdl__simple, faults=False)

    f = lambda r, s : client.service.f(__inject={"reply":suds.byte_str(r),

        "status":s})

    assert f("", None) is not None

    assert f("", http.client.INTERNAL_SERVER_ERROR) is not None

    assert f("", http.client.ACCEPTED) is None

    assert f("", http.client.NO_CONTENT) is None

    assert f("bla-bla", http.client.ACCEPTED) is None

    assert f("bla-bla", http.client.NO_CONTENT) is None





def test_badly_formed_reply_XML():

    for faults in (True, False):

        client = tests.client_from_wsdl(_wsdl__simple, faults=faults)

        pytest.raises(xml.sax.SAXParseException, client.service.f,

            __inject={"reply":suds.byte_str("bad food")})





# TODO: Update the current restriction type output parameter handling so such

# parameters get converted to the correct Python data type based on the

# restriction's underlying data type.

@pytest.mark.xfail

def test_restriction_data_types():

    client_unnamed = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="Elemento">

        <xsd:simpleType>

          <xsd:restriction base="xsd:int">

            <xsd:enumeration value="1" />

            <xsd:enumeration value="3" />

            <xsd:enumeration value="5" />

          </xsd:restriction>

        </xsd:simpleType>

      </xsd:element>""", "Elemento"))



    client_named = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:simpleType name="MyType">

        <xsd:restriction base="xsd:int">

          <xsd:enumeration value="1" />

          <xsd:enumeration value="3" />

          <xsd:enumeration value="5" />

        </xsd:restriction>

      </xsd:simpleType>

      <xsd:element name="Elemento" type="ns:MyType" />""", "Elemento"))



    client_twice_restricted = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:simpleType name="MyTypeGeneric">

        <xsd:restriction base="xsd:int">

          <xsd:enumeration value="1" />

          <xsd:enumeration value="2" />

          <xsd:enumeration value="3" />

          <xsd:enumeration value="4" />

          <xsd:enumeration value="5" />

        </xsd:restriction>

      </xsd:simpleType>

      <xsd:simpleType name="MyType">

        <xsd:restriction base="ns:MyTypeGeneric">

          <xsd:enumeration value="1" />

          <xsd:enumeration value="3" />

          <xsd:enumeration value="5" />

        </xsd:restriction>

      </xsd:simpleType>

      <xsd:element name="Elemento" type="ns:MyType" />""", "Elemento"))



    for client in (client_unnamed, client_named, client_twice_restricted):

        response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <Elemento xmlns="my-namespace">5</Elemento>

  </env:Body>

</env:Envelope>""")))

        assert response.__class__ is int

        assert response == 5





def test_builtin_data_types(monkeypatch):

    monkeypatch.delitem(locals(), "e", False)



    integer_type_mapping = {

        "byte":int,

        "int":int,

        "integer":int,

        "long":int,

        "negativeInteger":int,

        "nonNegativeInteger":int,

        "nonPositiveInteger":int,

        "positiveInteger":int,

        "short":int,

        "unsignedByte":int,

        "unsignedInt":int,

        "unsignedLong":int,

        "unsignedShort":int}

    for tag, type in list(integer_type_mapping.items()):

        client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="value" type="xsd:%s" />""" % tag, "value"))

        response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <value xmlns="my-namespace">15</value>

  </env:Body>

</env:Envelope>""")))

        assert response.__class__ is type

        assert response == 15



    boolean_mapping = {"0":False, "1":True, "false":False, "true":True}

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="value" type="xsd:boolean" />""", "value"))

    for value, expected_value in list(boolean_mapping.items()):

        response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <value xmlns="my-namespace">%s</value>

  </env:Body>

</env:Envelope>""" % value)))

        assert response.__class__ is bool

        assert response == expected_value



    # Suds implements no extra range checking.

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="value" type="xsd:byte" />""", "value"))

    response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <value xmlns="my-namespace">1500</value>

  </env:Body>

</env:Envelope>""")))

    assert response.__class__ is int

    assert response == 1500



    #   Suds raises raw Python exceptions when it fails to convert received

    # response element data to its mapped Python integer data type, according

    # to the used WSDL schema.

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="value" type="xsd:int" />""", "value"))

    e = pytest.raises(ValueError, client.service.f, __inject=dict(

        reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <value xmlns="my-namespace">Fifteen</value>

  </env:Body>

</env:Envelope>"""))).value

    # ValueError instance received here has different string representations

    # depending on the Python version used:

    #   Python 2.4:

    #     "invalid literal for int(): Fifteen"

    #   Python 2.7.3, 3.2.3:

    #     "invalid literal for int() with base 10: 'Fifteen'"

    assert re.match("invalid literal for int\(\)( with base 10)?: ('?)Fifteen"

        "\\2$", str(e))



    # Suds returns invalid boolean values as None.

    invalid_boolean_values = ("True", "", "False", "2", "Fedora", "Z", "-1")

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="value" type="xsd:boolean" />""", "value"))

    for value in invalid_boolean_values:

        response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <value xmlns="my-namespace">%s</value>

  </env:Body>

</env:Envelope>""" % value)))

        assert response is None





def test_disabling_automated_simple_interface_unwrapping():

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="Wrapper">

        <xsd:complexType>

          <xsd:sequence>

            <xsd:element name="Elemento" type="xsd:string" />

          </xsd:sequence>

        </xsd:complexType>

      </xsd:element>""", "Wrapper"), unwrap=False)

    assert not _isOutputWrapped(client, "f")



    response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <Wrapper xmlns="my-namespace">

        <Elemento>La-di-da-da-da</Elemento>

    </Wrapper>

  </env:Body>

</env:Envelope>""")))



    assert response.__class__.__name__ == "Wrapper"

    assert len(response.__class__.__bases__) == 1

    assert response.__class__.__bases__[0] is suds.sudsobject.Object

    assert response.Elemento.__class__ is suds.sax.text.Text

    assert response.Elemento == "La-di-da-da-da"





def test_empty_reply():

    client = tests.client_from_wsdl(_wsdl__simple, faults=False)

    f = lambda status=None, description=None : client.service.f(__inject=dict(

        reply=suds.byte_str(), status=status, description=description))

    status, reason = f()

    assert status == http.client.OK

    assert reason is None

    status, reason = f(http.client.OK)

    assert status == http.client.OK

    assert reason is None

    status, reason = f(http.client.INTERNAL_SERVER_ERROR)

    assert status == http.client.INTERNAL_SERVER_ERROR

    assert reason == 'injected reply'

    status, reason = f(http.client.FORBIDDEN)

    assert status == http.client.FORBIDDEN

    assert reason == 'injected reply'

    status, reason = f(http.client.FORBIDDEN, "kwack")

    assert status == http.client.FORBIDDEN

    assert reason == 'kwack'





def test_fault_reply_with_unicode_faultstring(monkeypatch):

    monkeypatch.delitem(locals(), "e", False)



    unicode_string = "€ Jurko Gospodnetić ČĆŽŠĐčćžšđ"

    fault_xml = suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <env:Fault>

      <faultcode>env:Client</faultcode>

      <faultstring>%s</faultstring>

    </env:Fault>

  </env:Body>

</env:Envelope>

""" % unicode_string)



    client = tests.client_from_wsdl(_wsdl__simple, faults=True)

    inject = dict(reply=fault_xml, status=http.client.INTERNAL_SERVER_ERROR)

    e = pytest.raises(suds.WebFault, client.service.f, __inject=inject).value

    assert e.fault.faultstring == unicode_string

    assert e.document.__class__ is suds.sax.document.Document



    client = tests.client_from_wsdl(_wsdl__simple, faults=False)

    status, fault = client.service.f(__inject=dict(reply=fault_xml,

        status=http.client.INTERNAL_SERVER_ERROR))

    assert status == http.client.INTERNAL_SERVER_ERROR

    assert fault.faultstring == unicode_string





def test_invalid_fault_namespace(monkeypatch):

    monkeypatch.delitem(locals(), "e", False)



    fault_xml = suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/" xmlns:p="x">

  <env:Body>

    <p:Fault>

      <faultcode>env:Client</faultcode>

      <faultstring>Dummy error.</faultstring>

      <detail>

        <errorcode>ultimate</errorcode>

      </detail>

    </p:Fault>

  </env:Body>

</env:Envelope>

""")

    client = tests.client_from_wsdl(_wsdl__simple, faults=False)

    inject = dict(reply=fault_xml, status=http.client.OK)

    e = pytest.raises(Exception, client.service.f, __inject=inject).value

    assert e.__class__ is Exception

    assert str(e) == "<faultcode/> not mapped to message part"



    for http_status in (http.client.INTERNAL_SERVER_ERROR,

        http.client.PAYMENT_REQUIRED):

        status, reason = client.service.f(__inject=dict(reply=fault_xml,

            status=http_status, description="trla baba lan"))

        assert status == http_status

        assert reason == "trla baba lan"





def test_missing_wrapper_response():

    """

    Suds library's automatic structure unwrapping should not be applied to

    interpreting received SOAP Response XML.



    """

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="Wrapper">

        <xsd:complexType>

          <xsd:sequence>

            <xsd:element name="fResponse" type="xsd:string" />

          </xsd:sequence>

        </xsd:complexType>

      </xsd:element>""", "Wrapper"))

    assert _isOutputWrapped(client, "f")



    response_with_missing_wrapper = client.service.f(__inject=dict(

        reply=suds.byte_str("""<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <fResponse xmlns="my-namespace">Anything</fResponse>

  </env:Body>

</env:Envelope>""")))

    assert response_with_missing_wrapper is None





def test_reply_error_with_detail_with_fault(monkeypatch):

    monkeypatch.delitem(locals(), "e", False)



    client = tests.client_from_wsdl(_wsdl__simple, faults=True)



    for http_status in (http.client.OK, http.client.INTERNAL_SERVER_ERROR):

        inject = dict(reply=_fault_reply__with_detail, status=http_status)

        e = pytest.raises(suds.WebFault, client.service.f, __inject=inject)

        e = e.value

        _test_fault(e.fault, True)

        assert e.document.__class__ is suds.sax.document.Document

        assert str(e) == "Server raised fault: 'Dummy error.'"



    inject = dict(reply=_fault_reply__with_detail, status=http.client.BAD_REQUEST,

        description="quack-quack")

    e = pytest.raises(Exception, client.service.f, __inject=inject).value

    assert e.__class__ is Exception

    assert e.args[0][0] == http.client.BAD_REQUEST

    assert e.args[0][1] == "quack-quack"





def test_reply_error_with_detail_without_fault():

    client = tests.client_from_wsdl(_wsdl__simple, faults=False)



    for http_status in (http.client.OK, http.client.INTERNAL_SERVER_ERROR):

        status, fault = client.service.f(__inject=dict(

            reply=_fault_reply__with_detail, status=http_status))

        assert status == http.client.INTERNAL_SERVER_ERROR

        _test_fault(fault, True)



    status, fault = client.service.f(__inject=dict(

        reply=_fault_reply__with_detail, status=http.client.BAD_REQUEST))

    assert status == http.client.BAD_REQUEST

    assert fault == "injected reply"



    status, fault = client.service.f(__inject=dict(

        reply=_fault_reply__with_detail, status=http.client.BAD_REQUEST,

        description="haleluja"))

    assert status == http.client.BAD_REQUEST

    assert fault == "haleluja"





def test_reply_error_without_detail_with_fault(monkeypatch):

    monkeypatch.delitem(locals(), "e", False)



    client = tests.client_from_wsdl(_wsdl__simple, faults=True)



    for http_status in (http.client.OK, http.client.INTERNAL_SERVER_ERROR):

        inject = dict(reply=_fault_reply__without_detail, status=http_status)

        e = pytest.raises(suds.WebFault, client.service.f, __inject=inject)

        e = e.value

        _test_fault(e.fault, False)

        assert e.document.__class__ is suds.sax.document.Document

        assert str(e) == "Server raised fault: 'Dummy error.'"



    inject = dict(reply=_fault_reply__with_detail, status=http.client.BAD_REQUEST,

        description="quack-quack")

    e = pytest.raises(Exception, client.service.f, __inject=inject).value

    assert e.__class__ is Exception

    assert e.args[0][0] == http.client.BAD_REQUEST

    assert e.args[0][1] == "quack-quack"





def test_reply_error_without_detail_without_fault():

    client = tests.client_from_wsdl(_wsdl__simple, faults=False)



    for http_status in (http.client.OK, http.client.INTERNAL_SERVER_ERROR):

        status, fault = client.service.f(__inject=dict(

            reply=_fault_reply__without_detail, status=http_status))

        assert status == http.client.INTERNAL_SERVER_ERROR

        _test_fault(fault, False)



    status, fault = client.service.f(__inject=dict(

        reply=_fault_reply__without_detail, status=http.client.BAD_REQUEST,

        description="kung-fu-fui"))

    assert status == http.client.BAD_REQUEST

    assert fault == "kung-fu-fui"





def test_simple_bare_and_wrapped_output():

    # Prepare web service proxies.

    client_bare = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="fResponse" type="xsd:string" />""", "fResponse"))

    client_wrapped = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="Wrapper">

        <xsd:complexType>

          <xsd:sequence>

            <xsd:element name="fResponse" type="xsd:string" />

          </xsd:sequence>

        </xsd:complexType>

      </xsd:element>""", "Wrapper"))



    #   Make sure suds library inteprets our WSDL definitions as wrapped or

    # bare output interfaces as expected.

    assert not _isOutputWrapped(client_bare, "f")

    assert _isOutputWrapped(client_wrapped, "f")



    #   Both bare & wrapped single parameter output web service operation

    # results get presented the same way even though the wrapped one actually

    # has an extra wrapper element around its received output data.

    data = "The meaning of life."

    get_response = lambda client, x : client.service.f(__inject=dict(

        reply=suds.byte_str(x)))



    response_bare = get_response(client_bare, """<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <fResponse xmlns="my-namespace">%s</fResponse>

  </env:Body>

</env:Envelope>""" % data)

    assert response_bare.__class__ is suds.sax.text.Text

    assert response_bare == data



    response_wrapped = get_response(client_wrapped, """<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <Wrapper xmlns="my-namespace">

      <fResponse>%s</fResponse>

    </Wrapper>

  </env:Body>

</env:Envelope>""" % data)

    assert response_wrapped.__class__ is suds.sax.text.Text

    assert response_wrapped == data





def test_wrapped_sequence_output():

    client = tests.client_from_wsdl(tests.wsdl_output("""\

      <xsd:element name="Wrapper">

        <xsd:complexType>

          <xsd:sequence>

            <xsd:element name="result1" type="xsd:string" />

            <xsd:element name="result2" type="xsd:string" />

            <xsd:element name="result3" type="xsd:string" />

          </xsd:sequence>

        </xsd:complexType>

      </xsd:element>""", "Wrapper"))

    assert _isOutputWrapped(client, "f")



    response = client.service.f(__inject=dict(reply=suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <Wrapper xmlns="my-namespace">

        <result1>Uno</result1>

        <result2>Due</result2>

        <result3>Tre</result3>

    </Wrapper>

  </env:Body>

</env:Envelope>""")))



    #   Composite replies always get unmarshalled as a dynamically constructed

    # class named 'reply'.

    assert len(response.__class__.__bases__) == 1

    assert response.__class__.__name__ == "reply"

    assert response.__class__.__bases__[0] is suds.sudsobject.Object



    # Check response content.

    assert len(response) == 3

    assert response.result1 == "Uno"

    assert response.result2 == "Due"

    assert response.result3 == "Tre"

    assert response.result1.__class__ is suds.sax.text.Text

    assert response.result2.__class__ is suds.sax.text.Text

    assert response.result3.__class__ is suds.sax.text.Text





def _attibutes(object):

    result = set()

    for x in object:

        result.add(x[0])

    return result





def _isOutputWrapped(client, method_name):

    assert len(client.wsdl.bindings) == 1

    operation = list(client.wsdl.bindings.values())[0].operations[method_name]

    return operation.soap.output.body.wrapped





def _test_fault(fault, has_detail):

    assert fault.faultcode == "env:Client"

    assert fault.faultstring == "Dummy error."

    assert hasattr(fault, "detail") == has_detail

    assert not has_detail or fault.detail.errorcode == "ultimate"

    assert not hasattr(fault, "nonexisting")

    expected_attributes = set(("faultcode", "faultstring"))

    if has_detail:

        expected_attributes.add("detail")

    assert _attibutes(fault) == expected_attributes

    assert not has_detail or _attibutes(fault.detail) == set(("errorcode",))





_fault_reply__with_detail = suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <env:Fault>

      <faultcode>env:Client</faultcode>

      <faultstring>Dummy error.</faultstring>

      <detail>

        <errorcode>ultimate</errorcode>

      </detail>

    </env:Fault>

  </env:Body>

</env:Envelope>

""")



_fault_reply__without_detail = suds.byte_str("""\

<?xml version="1.0"?>

<env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">

  <env:Body>

    <env:Fault>

      <faultcode>env:Client</faultcode>

      <faultstring>Dummy error.</faultstring>

    </env:Fault>

  </env:Body>

</env:Envelope>

""")



_wsdl__simple = tests.wsdl_output("""\

      <xsd:element name="fResponse">

        <xsd:complexType>

          <xsd:sequence>

            <xsd:element name="output_i" type="xsd:integer" />

            <xsd:element name="output_s" type="xsd:string" />

          </xsd:sequence>

        </xsd:complexType>

      </xsd:element>""", "fResponse")




# -*- coding: utf-8 -*-



# This program is free software; you can redistribute it and/or modify

# it under the terms of the (LGPL) GNU Lesser General Public License as

# published by the Free Software Foundation; either version 3 of the

# License, or (at your option) any later version.

#

# This program is distributed in the hope that it will be useful,

# but WITHOUT ANY WARRANTY; without even the implied warranty of

# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

# GNU Library Lesser General Public License for more details at

# ( http://www.gnu.org/licenses/lgpl.html ).

#

# You should have received a copy of the GNU Lesser General Public License

# along with this program; if not, write to the Free Software

# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )



"""

Suds Python library document caching unit tests.



Implemented using the 'pytest' testing framework.



"""



if __name__ == "__main__":

    from . import __init__

    __init__.runUsingPyTest(globals())





import suds

import suds.cache

import suds.sax.parser



import pytest



import os

import tempfile





class InvisibleMan:

    """Dummy class used for pickling related tests."""

    def __init__(self, x):

        self.x = x





# Hardcoded values used in different caching test cases.

value_empty = suds.byte_str("")

value_f2 = suds.byte_str("fifi2")

value_f22 = suds.byte_str("fifi22")

value_f3 = suds.byte_str("fifi3")

value_p1 = suds.byte_str("pero1")

value_p11 = suds.byte_str("pero11")

value_p111 = suds.byte_str("pero111")

value_p2 = suds.byte_str("pero2")

value_p22 = suds.byte_str("pero22")

value_unicode = suds.byte_str("€ 的 čćžšđČĆŽŠĐ")





def test_Cache():

    cache = suds.cache.Cache()

    pytest.raises(Exception, cache.get, "id")

    pytest.raises(Exception, cache.put, "id", "object")

    pytest.raises(Exception, cache.purge, "id")

    pytest.raises(Exception, cache.clear)





def test_DocumentCache(tmpdir):

    cacheFolder = tmpdir.join("puffy").strpath

    cache = suds.cache.DocumentCache(cacheFolder)

    assert isinstance(cache, suds.cache.FileCache)

    assert cache.get("unga1") is None



    # TODO: DocumentCache class interface seems silly. Its get() operation

    # returns an XML document while its put() operation takes an XML element.

    # The put() operation also silently ignores passed data of incorrect type.

    # TODO: Update this test to no longer depend on the exact input XML data

    # formatting. We currently expect it to be formatted exactly as what gets

    # read back from the DocumentCache.

    content = suds.byte_str("""\

<xsd:element name="Elemento">

   <xsd:simpleType>

      <xsd:restriction base="xsd:string">

         <xsd:enumeration value="alfa"/>

         <xsd:enumeration value="beta"/>

         <xsd:enumeration value="gamma"/>

      </xsd:restriction>

   </xsd:simpleType>

</xsd:element>""")

    xml = suds.sax.parser.Parser().parse(suds.BytesIO(content))

    cache.put("unga1", xml.getChildren()[0])

    readXML = cache.get("unga1")

    assert isinstance(readXML, suds.sax.document.Document)

    readXMLElements = readXML.getChildren()

    assert len(readXMLElements) == 1

    readXMLElement = readXMLElements[0]

    assert isinstance(readXMLElement, suds.sax.element.Element)

    assert suds.byte_str(str(readXMLElement)) == content





def test_FileCache():

    cache = suds.cache.FileCache()

    assert isinstance(cache, suds.cache.Cache)





def test_FileCache_clear(tmpdir):

    cacheFolder1 = tmpdir.join("fungus").strpath

    cache1 = suds.cache.FileCache(cacheFolder1)

    cache1.put("unga1", value_p1)

    cache1.put("unga2", value_p2)

    assert cache1.get("unga1") == value_p1

    assert cache1.get("unga2") == value_p2

    cache1.clear()

    assert _isEmptyCacheFolder(cacheFolder1)

    assert cache1.get("unga1") is None

    assert cache1.get("unga2") is None

    cache1.put("unga1", value_p11)

    cache1.put("unga2", value_p2)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2



    cacheFolder2 = tmpdir.join("broccoli").strpath

    cache2 = suds.cache.FileCache(cacheFolder2)

    cache2.put("unga2", value_f2)

    assert cache2.get("unga2") == value_f2

    cache2.clear()

    assert not _isEmptyCacheFolder(cacheFolder1)

    assert _isEmptyCacheFolder(cacheFolder2)

    assert cache2.get("unga2") is None

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2

    cache2.put("unga2", value_p22)

    assert cache2.get("unga2") == value_p22





def test_FileCache_location(tmpdir):

    defaultLocation = os.path.join(tempfile.gettempdir(), "suds")

    cache = suds.cache.FileCache()

    assert os.path.isdir(cache.location)

    assert cache.location == defaultLocation

    assert suds.cache.FileCache().location == defaultLocation

    assert cache.location == defaultLocation



    cacheFolder1 = tmpdir.join("flip-flop1").strpath

    assert not os.path.isdir(cacheFolder1)

    assert suds.cache.FileCache(location=cacheFolder1).location == cacheFolder1

    assert _isEmptyCacheFolder(cacheFolder1)



    cacheFolder2 = tmpdir.join("flip-flop2").strpath

    assert not os.path.isdir(cacheFolder2)

    assert suds.cache.FileCache(cacheFolder2).location == cacheFolder2

    assert _isEmptyCacheFolder(cacheFolder2)





def test_FileCache_close_leaves_cached_files_behind(tmpdir):

    cacheFolder1 = tmpdir.join("ana").strpath

    cache1 = suds.cache.FileCache(cacheFolder1)

    cache1.put("unga1", value_p1)

    cache1.put("unga2", value_p2)



    cacheFolder2 = tmpdir.join("nan").strpath

    cache2 = suds.cache.FileCache(cacheFolder2)

    cache2.put("unga2", value_f2)

    cache2.put("unga3", value_f3)



    del cache1



    cache11 = suds.cache.FileCache(cacheFolder1)

    assert cache11.get("unga1") == value_p1

    assert cache11.get("unga2") == value_p2

    assert cache2.get("unga2") == value_f2

    assert cache2.get("unga3") == value_f3





def test_FileCache_get_put(tmpdir):

    cacheFolder1 = tmpdir.join("firefly").strpath

    cache1 = suds.cache.FileCache(cacheFolder1)

    assert _isEmptyCacheFolder(cacheFolder1)

    assert cache1.get("unga1") is None

    cache1.put("unga1", value_p1)

    assert not _isEmptyCacheFolder(cacheFolder1)

    assert cache1.get("unga1") == value_p1

    assert cache1.get("unga2") is None

    cache1.put("unga1", value_p11)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") is None

    cache1.put("unga2", value_p2)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2



    cacheFolder2 = tmpdir.join("semper fi").strpath

    cache2 = suds.cache.FileCache(cacheFolder2)

    assert _isEmptyCacheFolder(cacheFolder2)

    assert cache2.get("unga2") is None

    cache2.put("unga2", value_f2)

    assert not _isEmptyCacheFolder(cacheFolder2)

    assert cache2.get("unga2") == value_f2

    assert cache2.get("unga3") is None

    cache2.put("unga2", value_f22)

    assert cache2.get("unga2") == value_f22

    assert cache2.get("unga3") is None

    cache2.put("unga3", value_f3)

    assert cache2.get("unga2") == value_f22

    assert cache2.get("unga3") == value_f3



    assert not _isEmptyCacheFolder(cacheFolder1)

    assert not _isEmptyCacheFolder(cacheFolder2)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2

    assert cache1.get("unga3") is None

    assert cache2.get("unga1") is None

    assert cache2.get("unga2") == value_f22

    assert cache2.get("unga3") == value_f3





def test_FileCache_purge(tmpdir):

    cacheFolder1 = tmpdir.join("flamenco").strpath

    cache1 = suds.cache.FileCache(cacheFolder1)

    cache1.put("unga1", value_p1)

    assert cache1.get("unga1") == value_p1

    cache1.purge("unga1")

    assert _isEmptyCacheFolder(cacheFolder1)

    assert cache1.get("unga1") is None

    cache1.put("unga1", value_p11)

    cache1.put("unga2", value_p2)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2

    cache1.purge("unga1")

    assert cache1.get("unga1") is None

    assert cache1.get("unga2") == value_p2

    cache1.put("unga1", value_p111)



    cacheFolder2 = tmpdir.join("shadow").strpath

    cache2 = suds.cache.FileCache(cacheFolder2)

    cache2.put("unga2", value_f2)

    cache2.purge("unga2")

    assert _isEmptyCacheFolder(cacheFolder2)

    assert cache1.get("unga1") == value_p111

    assert cache1.get("unga2") == value_p2

    assert cache2.get("unga2") is None





def test_FileCache_reused_cache_folder(tmpdir):

    cacheFolder = tmpdir.strpath

    cache1 = suds.cache.FileCache(cacheFolder)

    assert _isEmptyCacheFolder(cacheFolder)

    assert cache1.get("unga1") is None

    cache1.put("unga1", value_p1)

    assert cache1.get("unga1") == value_p1

    assert cache1.get("unga2") is None

    cache1.put("unga1", value_p11)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") is None

    cache1.put("unga2", value_p2)

    assert cache1.get("unga1") == value_p11

    assert cache1.get("unga2") == value_p2



    cache2 = suds.cache.FileCache(cacheFolder)

    assert cache2.get("unga1") == value_p11

    assert cache2.get("unga2") == value_p2

    cache2.put("unga3", value_f3)

    assert cache1.get("unga3") == value_f3





def test_FileCache_version(tmpdir):

    fakeVersionInfo = "--- fake version info ---"

    assert suds.__version__ != fakeVersionInfo



    cacheFolder = tmpdir.join("hitori")

    versionFile = cacheFolder.join("version")

    cache = suds.cache.FileCache(cacheFolder.strpath)

    assert versionFile.read() == suds.__version__

    cache.put("unga1", value_p1)



    versionFile.write(fakeVersionInfo)

    assert cache.get("unga1") == value_p1



    cache2 = suds.cache.FileCache(cacheFolder.strpath)

    assert _isEmptyCacheFolder(cacheFolder.strpath)

    assert cache.get("unga1") is None

    assert cache2.get("unga1") is None

    assert versionFile.read() == suds.__version__

    cache.put("unga1", value_p11)

    cache.put("unga2", value_p22)



    versionFile.remove()

    assert cache.get("unga1") == value_p11

    assert cache.get("unga2") == value_p22



    cache3 = suds.cache.FileCache(cacheFolder.strpath)

    assert _isEmptyCacheFolder(cacheFolder.strpath)

    assert cache.get("unga1") is None

    assert cache.get("unga2") is None

    assert cache2.get("unga1") is None

    assert versionFile.read() == suds.__version__





def test_FileCache_with_empty_cached_content(tmpdir):

    cacheFolder = tmpdir.strpath

    cache = suds.cache.FileCache(cacheFolder)

    cache.put("unga1", value_empty)

    assert cache.get("unga1") == value_empty

    assert not _isEmptyCacheFolder(cacheFolder)





def test_FileCache_with_random_utf_character_cached_content(tmpdir):

    cacheFolder = tmpdir.strpath

    cache = suds.cache.FileCache(cacheFolder)

    cache.put("unga1", value_unicode)

    assert cache.get("unga1") == value_unicode

    assert not _isEmptyCacheFolder(cacheFolder)





def test_NoCache():

    cache = suds.cache.NoCache()

    assert isinstance(cache, suds.cache.Cache)



    assert cache.get("id") == None

    cache.put("id", "something")

    assert cache.get("id") == None



    # TODO: It should not be an error to call purge() or clear() on a NoCache

    # instance.

    pytest.raises(Exception, cache.purge, "id")

    pytest.raises(Exception, cache.clear)





def test_ObjectCache(tmpdir):

    cacheFolder = tmpdir.join("george carlin").strpath

    cache = suds.cache.ObjectCache(cacheFolder)

    assert isinstance(cache, suds.cache.FileCache)

    assert cache.get("unga1") is None

    assert cache.get("unga2") is None

    cache.put("unga1", InvisibleMan(1))

    cache.put("unga2", InvisibleMan(2))

    read1 = cache.get("unga1")

    read2 = cache.get("unga2")

    assert read1.__class__ is InvisibleMan

    assert read2.__class__ is InvisibleMan

    assert read1.x == 1

    assert read2.x == 2





def _isEmptyCacheFolder(folder):

    assert os.path.isdir(folder)

    def walkError(error):

        pytest.fail("Error attempting to walk through cache folder contents.")

    count = 0

    for root, folders, files in os.walk(folder, onerror=walkError):

        assert root == folder

        return len(folders) == 0 and len(files) == 1 and files[0] == 'version'

    return False




# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

#
# sax encoding/decoding test.
#

from suds.sax.element import Element
from suds.sax.parser import Parser

def basic():
    xml = "<a>Me &amp;&amp; &lt;b&gt;my&lt;/b&gt; shadow&apos;s &lt;i&gt;dog&lt;/i&gt; love to &apos;play&apos; and sing &quot;la,la,la&quot;;</a>"
    p = Parser()
    d = p.parse(string=xml)
    a = d.root()
    print('A(parsed)=\n%s' % a)
    assert str(a) == xml
    b = Element('a')
    b.setText('Me &&amp; &lt;b>my</b> shadow\'s <i>dog</i> love to \'play\' and sing "la,la,la";')
    print('B(encoded)=\n%s' % b)
    assert str(b) == xml
    print('A(text-decoded)=\n%s' % a.getText())
    print('B(text-decoded)=\n%s' % b.getText())
    assert a.getText() == b.getText()
    print('test pruning')
    j = Element('A')
    j.set('n', 1)
    j.append(Element('B'))
    print(j)
    j.prune()
    print(j)

def cdata():
    xml = '<a><![CDATA[<b>This is my &amp;&lt;tag&gt;</b>]]></a>'
    p = Parser()
    d = p.parse(string=xml)
    print(d)
    a = d.root()
    print(a.getText())

if __name__ == '__main__':
    #basic()
    cdata()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
The I{sxbase} module provides I{base} classes representing schema objects.
"""

from suds import *
from suds.xsd import *
from suds.sax.element import Element
from suds.sax import Namespace

from logging import getLogger
log = getLogger(__name__)


class SchemaObject(UnicodeMixin):
    """
    A schema object is an extension to object with schema awareness.
    @ivar root: The XML root element.
    @type root: L{Element}
    @ivar schema: The schema containing this object.
    @type schema: L{schema.Schema}
    @ivar form_qualified: A flag that indicates that @elementFormDefault
        has a value of I{qualified}.
    @type form_qualified: boolean
    @ivar nillable: A flag that indicates that @nillable
        has a value of I{true}.
    @type nillable: boolean
    @ivar default: The default value.
    @type default: object
    @ivar rawchildren: A list raw of all children.
    @type rawchildren: [L{SchemaObject},...]
    """

    @classmethod
    def prepend(cls, d, s, filter=Filter()):
        """
        Prepend schema objects from B{s}ource list to
        the B{d}estination list while applying the filter.
        @param d: The destination list.
        @type d: list
        @param s: The source list.
        @type s: list
        @param filter: A filter that allows items to be prepended.
        @type filter: L{Filter}
        """
        i = 0
        for x in s:
            if x in filter:
                d.insert(i, x)
                i += 1

    @classmethod
    def append(cls, d, s, filter=Filter()):
        """
        Append schema objects from B{s}ource list to
        the B{d}estination list while applying the filter.
        @param d: The destination list.
        @type d: list
        @param s: The source list.
        @type s: list
        @param filter: A filter that allows items to be appended.
        @type filter: L{Filter}
        """
        for item in s:
            if item in filter:
                d.append(item)

    def __init__(self, schema, root):
        """
        @param schema: The containing schema.
        @type schema: L{schema.Schema}
        @param root: The XML root node.
        @type root: L{Element}
        """
        self.schema = schema
        self.root = root
        self.id = objid(self)
        self.name = root.get('name')
        self.qname = (self.name, schema.tns[1])
        self.min = root.get('minOccurs')
        self.max = root.get('maxOccurs')
        self.type = root.get('type')
        self.ref = root.get('ref')
        self.form_qualified = schema.form_qualified
        self.nillable = False
        self.default = root.get('default')
        self.rawchildren = []

    def attributes(self, filter=Filter()):
        """
        Get only the attribute content.
        @param filter: A filter to constrain the result.
        @type filter: L{Filter}
        @return: A list of tuples (attr, ancestry)
        @rtype: [(L{SchemaObject}, [L{SchemaObject},..]),..]
        """
        result = []
        for child, ancestry in self:
            if child.isattr() and child in filter:
                result.append((child, ancestry))
        return result

    def children(self, filter=Filter()):
        """
        Get only the I{direct} or non-attribute content.
        @param filter: A filter to constrain the result.
        @type filter: L{Filter}
        @return: A list tuples: (child, ancestry)
        @rtype: [(L{SchemaObject}, [L{SchemaObject},..]),..]
        """
        result = []
        for child, ancestry in self:
            if not child.isattr() and child in filter:
                result.append((child, ancestry))
        return result

    def get_attribute(self, name):
        """
        Get (find) an attribute by name.
        @param name: A attribute name.
        @type name: str
        @return: A tuple: the requested (attribute, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])
        """
        for child, ancestry in self.attributes():
            if child.name == name:
                return child, ancestry
        return None, []

    def get_child(self, name):
        """
        Get (find) a I{non-attribute} child by name.
        @param name: A child name.
        @type name: str
        @return: A tuple: the requested (child, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])
        """
        for child, ancestry in self.children():
            if child.any() or child.name == name:
                return child, ancestry
        return None, []

    def namespace(self, prefix=None):
        """
        Get this property's namespace.
        @param prefix: The default prefix.
        @type prefix: str
        @return: The schema's target namespace
        @rtype: (I{prefix},I{URI})
        """
        ns = self.schema.tns
        if ns[0] is None:
            ns = (prefix, ns[1])
        return ns

    def default_namespace(self):
        return self.root.defaultNamespace()

    def multi_occurrence(self):
        """
        Get whether the node has multiple occurrences, i.e. is a I{collection}.
        @return: True if it has, False if it has 1 occurrence at most.
        @rtype: boolean
        """
        max = self.max
        if max is None:
            return False
        if max.isdigit():
            return int(max) > 1
        return max == 'unbounded'

    def optional(self):
        """
        Get whether this type is optional.
        @return: True if optional, else False
        @rtype: boolean
        """
        return self.min == '0'

    def required(self):
        """
        Get whether this type is required.
        @return: True if required, else False
        @rtype: boolean
        """
        return not self.optional()

    def resolve(self, nobuiltin=False):
        """
        Resolve the node's type reference and return the referenced type node.

        Only schema objects that actually support 'having a type' custom
        implement this interface while others simply resolve as themselves.
        @param nobuiltin: Flag indicating whether resolving to an external XSD
            builtin type should not be allowed.
        @return: The resolved (true) type.
        @rtype: L{SchemaObject}
        """
        return self

    def sequence(self):
        """
        Get whether this is a <xs:sequence/>.
        @return: True if <xs:sequence/>, else False
        @rtype: boolean
        """
        return False

    def xslist(self):
        """
        Get whether this is a <xs:list/>.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def all(self):
        """
        Get whether this is an <xs:all/>.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def choice(self):
        """
        Get whether this is a <xs:choice/>.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def any(self):
        """
        Get whether this is an <xs:any/>.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def builtin(self):
        """
        Get whether this is a schema-instance (xs) type.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def enum(self):
        """
        Get whether this is a simple-type containing an enumeration.
        @return: True if any, else False
        @rtype: boolean
        """
        return False

    def isattr(self):
        """
        Get whether the object is a schema I{attribute} definition.
        @return: True if an attribute, else False.
        @rtype: boolean
        """
        return False

    def extension(self):
        """
        Get whether the object is an extension of another type.
        @return: True if an extension, else False.
        @rtype: boolean
        """
        return False

    def restriction(self):
        """
        Get whether the object is an restriction of another type.
        @return: True if an restriction, else False.
        @rtype: boolean
        """
        return False

    def mixed(self):
        """
        Get whether this I{mixed} content.
        """
        return False

    def find(self, qref, classes=()):
        """
        Find a referenced type in self or children.
        @param qref: A qualified reference.
        @type qref: qref
        @param classes: A list of classes used to qualify the match.
        @type classes: [I{class},...]
        @return: The referenced type.
        @rtype: L{SchemaObject}
        @see: L{qualify()}
        """
        if not len(classes):
            classes = (self.__class__,)
        if self.qname == qref and self.__class__ in classes:
            return self
        for c in self.rawchildren:
            p = c.find(qref, classes)
            if p is not None:
                return p
        return None

    def translate(self, value, topython=True):
        """
        Translate a value (type) to/from a Python type.
        @param value: A value to translate.
        @return: The converted I{language} type.
        """
        return value

    def childtags(self):
        """
        Get a list of valid child tag names.
        @return: A list of child tag names.
        @rtype: [str,...]
        """
        return ()

    def dependencies(self):
        """
        Get a list of dependencies for dereferencing.
        @return: A merge dependency index and a list of dependencies.
        @rtype: (int, [L{SchemaObject},...])
        """
        return None, []

    def autoqualified(self):
        """
        The list of I{auto} qualified attribute values.
        Qualification means to convert values into I{qref}.
        @return: A list of attibute names.
        @rtype: list
        """
        return ['type', 'ref']

    def qualify(self):
        """
        Convert attribute values, that are references to other
        objects, into I{qref}.  Qualified using the default document namespace.
        Since many WSDLs are written improperly: when the document does
        not define its default namespace, the schema target namespace is used
        to qualify references.
        """
        defns = self.root.defaultNamespace()
        if Namespace.none(defns):
            defns = self.schema.tns
        for a in self.autoqualified():
            ref = getattr(self, a)
            if ref is None:
                continue
            if isqref(ref):
                continue
            qref = qualify(ref, self.root, defns)
            log.debug('%s, convert %s="%s" to %s', self.id, a, ref, qref)
            setattr(self, a, qref)

    def merge(self, other):
        """
        Merge another object as needed.
        """
        other.qualify()
        for n in ('name',
                  'qname',
                  'min',
                  'max',
                  'default',
                  'type',
                  'nillable',
                  'form_qualified'):
            if getattr(self, n) is not None:
                continue
            v = getattr(other, n)
            if v is None:
                continue
            setattr(self, n, v)

    def content(self, collection=None, filter=Filter(), history=None):
        """
        Get a I{flattened} list of this node's contents.
        @param collection: A list to fill.
        @type collection: list
        @param filter: A filter used to constrain the result.
        @type filter: L{Filter}
        @param history: The history list used to prevent cyclic dependency.
        @type history: list
        @return: The filled list.
        @rtype: list
        """
        if collection is None:
            collection = []
        if history is None:
            history = []
        if self in history:
            return collection
        history.append(self)
        if self in filter:
            collection.append(self)
        for c in self.rawchildren:
            c.content(collection, filter, history[:])
        return collection

    def str(self, indent=0, history=None):
        """
        Get a string representation of this object.
        @param indent: The indent.
        @type indent: int
        @return: A string.
        @rtype: str
        """
        if history is None:
            history = []
        if self in history:
            return '%s ...' % Repr(self)
        history.append(self)
        tab = '%*s'%(indent*3, '')
        result = ['%s<%s' % (tab, self.id)]
        for n in self.description():
            if not hasattr(self, n):
                continue
            v = getattr(self, n)
            if v is None:
                continue
            result.append(' %s="%s"' % (n, v))
        if len(self):
            result.append('>')
            for c in self.rawchildren:
                result.append('\n')
                result.append(c.str(indent+1, history[:]))
                if c.isattr():
                    result.append('@')
            result.append('\n%s' % tab)
            result.append('</%s>' % self.__class__.__name__)
        else:
            result.append(' />')
        return ''.join(result)

    def description(self):
        """
        Get the names used for str() and repr() description.
        @return:  A dictionary of relevant attributes.
        @rtype: [str,...]
        """
        return ()

    def __unicode__(self):
        return str(self.str())

    def __repr__(self):
        s = []
        s.append('<%s' % self.id)
        for n in self.description():
            if not hasattr(self, n):
                continue
            v = getattr(self, n)
            if v is None:
                continue
            s.append(' %s="%s"' % (n, v))
        s.append(' />')
        return ''.join(s)

    def __len__(self):
        n = 0
        for x in self: n += 1
        return n

    def __iter__(self):
        return Iter(self)

    def __getitem__(self, index):
        """Returns a contained schema object referenced by its 0-based index.

        Returns None if such an object does not exist.

        """
        i = 0
        for c in self:
            if i == index:
                return c
            i += 1


class Iter:
    """
    The content iterator - used to iterate the L{Content} children.  The
    iterator provides a I{view} of the children that is free of container
    elements such as <sequence/> and <choice/>.
    @ivar stack: A stack used to control nesting.
    @type stack: list
    """

    class Frame:
        """ A content iterator frame. """

        def __init__(self, sx):
            """
            @param sx: A schema object.
            @type sx: L{SchemaObject}
            """
            self.sx = sx
            self.items = sx.rawchildren
            self.index = 0

        def __next__(self):
            """
            Get the I{next} item in the frame's collection.
            @return: The next item or None
            @rtype: L{SchemaObject}
            """
            if self.index < len(self.items):
                result = self.items[self.index]
                self.index += 1
                return result

    def __init__(self, sx):
        """
        @param sx: A schema object.
        @type sx: L{SchemaObject}
        """
        self.stack = []
        self.push(sx)

    def push(self, sx):
        """
        Create a frame and push the specified object.
        @param sx: A schema object to push.
        @type sx: L{SchemaObject}
        """
        self.stack.append(Iter.Frame(sx))

    def pop(self):
        """
        Pop the I{top} frame.
        @return: The popped frame.
        @rtype: L{Frame}
        @raise StopIteration: when stack is empty.
        """
        if len(self.stack):
            return self.stack.pop()
        else:
            raise StopIteration()

    def top(self):
        """
        Get the I{top} frame.
        @return: The top frame.
        @rtype: L{Frame}
        @raise StopIteration: when stack is empty.
        """
        if len(self.stack):
            return self.stack[-1]
        else:
            raise StopIteration()

    def __next__(self):
        """
        Get the next item.
        @return: A tuple: the next (child, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])
        @raise StopIteration: A the end.
        """
        frame = self.top()
        while True:
            result = next(frame)
            if result is None:
                self.pop()
                return next(self)
            if isinstance(result, Content):
                ancestry = [f.sx for f in self.stack]
                return result, ancestry
            self.push(result)
            return next(self)

    def __iter__(self):
        return self


class XBuiltin(SchemaObject):
    """
    Represents an (XSD) schema <xs:*/> node.
    """

    def __init__(self, schema, name):
        """
        @param schema: The containing schema.
        @type schema: L{schema.Schema}
        """
        root = Element(name)
        SchemaObject.__init__(self, schema, root)
        self.name = name
        self.nillable = True

    def namespace(self, prefix=None):
        return Namespace.xsdns

    def builtin(self):
        return True


class Content(SchemaObject):
    """
    This class represents those schema objects that represent
    real XML document content.
    """
    pass


class NodeFinder:
    """
    Find nodes based on flexable criteria.  The I{matcher}
    may be any object that implements a match(n) method.
    @ivar matcher: An object used as criteria for match.
    @type matcher: I{any}.match(n)
    @ivar limit: Limit the number of matches.  0=unlimited.
    @type limit: int
    """
    def __init__(self, matcher, limit=0):
        """
        @param matcher: An object used as criteria for match.
        @type matcher: I{any}.match(n)
        @param limit: Limit the number of matches.  0=unlimited.
        @type limit: int
        """
        self.matcher = matcher
        self.limit = limit

    def find(self, node, list):
        """
        Traverse the tree looking for matches.
        @param node: A node to match on.
        @type node: L{SchemaObject}
        @param list: A list to fill.
        @type list: list
        """
        if self.matcher.match(node):
            list.append(node)
            self.limit -= 1
            if self.limit == 0:
                return
        for c in node.rawchildren:
            self.find(c, list)
        return self
